"""Implementation of divider tool."""
import itertools
import os
import re
import logging
from typing import TextIO

from .textfixer import TextFixerWorker

WORDS_PATTER = r'\w+'
PARAGRAPH = "\n"
SENTENCE_END_1 = ". "
SENTENCE_END_2 = ".\n"
SPACE = " "

WORDS_NUMBER_SHIFT = 0.2
LANG_PL = 'pl'

_log = logging.getLogger(__name__)


def utf8size(text: str) -> int:
    """Returns size of string in utf8 encoding.

    :param text: string to check
    :type text: str
    :return: size of string in utf8 encoding
    :rtype: int
    """
    return len(text.encode('utf8'))


class Divider:
    """Class implementing divider."""

    def __init__(self,
                 chars_nbr: int = 20000,
                 last_chunk_ratio: float = 0.5,
                 sep_diff_ratio: float = 10.0,
                 separators: list = [". ", "!", "?", "...", ".", ",", " "]
                 ):
        """Constructor.

        :param chars_nbr: number of characters in chunk, defaults to 20000
        :type chars_nbr: int
        :param last_chunk_ratio: ratio of last chunk size to chars_nbr,
            defaults to 0.5
        :type last_chunk_ratio: float
        :param sep_diff_ratio: ratio of separators differences to chars_nbr,
            defaults to 10.0
        :type sep_diff_ratio: float
        :param separators: list of separators, defaults to [". ", "!", "?",
            "...", ".", ",", " "]
        :type separators: list
        """
        data_path = os.path.join(os.path.dirname(__file__),
                                 'data_words/scrabble-polish-words.txt')
        self.text_fixer = TextFixerWorker(data_path)
        self.chars_nbr = chars_nbr
        self.last_chunk_ratio = last_chunk_ratio
        self.sep_diff_ratio = sep_diff_ratio
        self.separators = separators

    def divide_into_chunks(self, input_file, words_nbr=10, lang=LANG_PL,
                           fix_text=False):
        """Divides input stream into chunks.

        :param input_file: path to input file
        :type input_file: str
        :param words_nbr: maximum number of words in chunk, defaults to 10
        :type words_nbr: int
        :param lang: language of text, defaults to 'pl'
        :type lang: str
        :param fix_text: if text should be fixed, defaults to False
        :type fix_text: bool
        :yield: a pair of chunk str and progress
        :rtype: Iterator[tuple(str, double)]
        """
        if fix_text and lang != LANG_PL:
            _log.warning("Text fixing is available only for polish!")
            fix_text = False

        words_nbr_shift = int(words_nbr * WORDS_NUMBER_SHIFT)\
            if fix_text else 0

        input_size = self._get_input_size(input_file)
        progress = 0
        with open(input_file, "r") as text:
            chunk = text.read(self.chars_nbr)
            while chunk:
                if len(re.findall(WORDS_PATTER, chunk))\
                        > words_nbr + words_nbr_shift:

                    split_idx = self._find_split_idx(chunk,
                                                     words_nbr +
                                                     words_nbr_shift)
                    text_to_split = self._fix_text(chunk[:split_idx])\
                        if fix_text else chunk[:split_idx]
                    cut_idx = self._check_paragraph(text_to_split, words_nbr)
                    output = text_to_split[:cut_idx] \
                        if split_idx != cut_idx else text_to_split
                    shift = len(chunk[:split_idx]) - len(text_to_split)
                    index = cut_idx + shift
                    # fix for one letter words
                    if output[-1] != chunk[index - 1] and \
                       chunk[index].isspace() and \
                       chunk[index - 1].isalpha() and \
                       chunk[index - 2].isspace():
                        index -= 1

                    if chunk[index].isalnum():
                        for _ in range(shift):
                            if chunk[index - 1].isalnum():
                                index -= 1
                                continue
                            break

                    chunk_tail = chunk[index:]

                    progress = self._get_progress(progress, input_size, output)

                    yield output, progress
                else:
                    output = self._fix_text(chunk)\
                        if fix_text else chunk
                    chunk_tail = ''
                    progress = self._get_progress(progress, input_size, output)
                    yield output, progress

                if len(chunk_tail) < 1:
                    chunk = text.read(self.chars_nbr)
                else:
                    chunk = chunk_tail

    def divide_into_paragraphs(
            self,
            input_file: TextIO,
            file_size: int,
    ):
        """Divides input stream into paragraphs.

        :param input_file: input file stream
        :type input_file: TextIO
        :param file_size: size of input file
        :type file_size: int
        :yield: a pair of paragraph str and progress
        :rtype: Iterator[tuple(str, double)]
        """
        processed, chunk_data = 0, ""
        while processed < file_size:
            buffer = input_file.read(self.chars_nbr)
            buffer_size = utf8size(buffer)
            processed += buffer_size
            if buffer_size == 0:
                if chunk_data:
                    if utf8size(chunk_data) <= self.last_chunk_ratio \
                            * self.chars_nbr:
                        processed -= utf8size(chunk_data)
                    yield chunk_data, processed / file_size
                processed = file_size

            chunk_data += buffer
            if utf8size(chunk_data) > self.chars_nbr:
                save_part, tail = self._divided_by_separator(chunk_data)
                if save_part:
                    yield save_part, processed / file_size
                    chunk_data = tail
                else:
                    yield chunk_data, processed / file_size
                    chunk_data = ""
            else:
                yield chunk_data, processed / file_size
                chunk_data = ""

    def _divided_by_separator(self,
                              chunk_data: str
                              ):
        for separator in self.separators:
            if separator[::-1] in chunk_data[::-1]:
                div_index = chunk_data.rindex(separator) + 1
                save_part_size = utf8size(chunk_data[:div_index])
                if (1 - self.sep_diff_ratio) * self.chars_nbr <= \
                        save_part_size \
                        <= (1 + self.sep_diff_ratio) * self.chars_nbr:
                    save_part = chunk_data[:div_index]
                    tail = chunk_data[div_index:]
                    return save_part, tail
        return None, chunk_data

    @staticmethod
    def _get_input_size(path_to_file):
        return os.stat(path_to_file).st_size

    @staticmethod
    def _get_progress(progress, input_size, output):
        return (progress * input_size + len(output)) / input_size

    @staticmethod
    def _find_split_idx(text, words_nbr):
        """Finds split index of string where defined number of words ends.

        :param text: value of text to check
        :type text: str
        :param words_nbr: number of words to set split index on
        :type words_nbr: int
        :return: split index for given words number
        :rtype: int
        """
        matches = re.finditer(WORDS_PATTER, text, flags=re.IGNORECASE)
        match = next(itertools.islice(matches, words_nbr - 1, None))
        return match.end()

    def _check_paragraph(self, chunk, words_nbr):
        """Checks if chunks contains end of paragraph in its second half.

        :param chunk: value of chunk to check
        :type chunk: str
        :param words_nbr: maximum number of words in chunk
        :type words_nbr: int
        :return: a pair of extracted chunk and chunk_tail
        :rtype: int
        """
        cut_conditions = [PARAGRAPH, SENTENCE_END_1, SENTENCE_END_2]
        cut_indexes = []
        for idx, cut_condition in enumerate(cut_conditions):
            if cut_condition in chunk:
                cut_indexes.append((chunk.rindex(cut_condition), idx))

        words_nbr_idx = len(chunk)
        if len(re.findall(WORDS_PATTER, chunk)) > words_nbr:
            words_nbr_idx = self._find_split_idx(chunk, words_nbr)

        cut_indexes = [(idx, cut_idx) for idx, cut_idx in cut_indexes
                       if idx <= words_nbr_idx]
        if len(cut_indexes) > 0:
            cut_indexes.sort(reverse=True, key=lambda x: x[0])
            idx, cut_condition_idx = cut_indexes[0]
            if idx > (len(chunk) / 2):
                return idx + len(cut_conditions[cut_condition_idx])
        return words_nbr_idx

    def _fix_text(self, text):
        return self.text_fixer.word_transfer(text.split('\n'))
